import logging

import psycopg2
from tc_hivemind_backend.db.credentials import load_postgres_credentials


class PostgresSingleton:
    _instance = None

    def __new__(cls, dbname: str | None, *args, **kwargs):
        """
        if `db_name` was `None`, connect to default database in .env
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance.connect(dbname)
        return cls._instance

    def connect(self, dbname: str | None):
        """
        connect to postgrsql instance
        """
        creds = load_postgres_credentials()
        try:
            self.conn = psycopg2.connect(
                dbname=dbname or creds["db_name"],
                user=creds["user"],
                password=creds["password"],
                host=creds["host"],
                port=creds["port"],
            )
        except psycopg2.OperationalError as exp:
            logging.error(f"Error initializing connection, exp: {exp}")
            self.destroy_instance()

    def get_connection(self):
        """
        get a connection to database
        """
        return self.conn

    def close_connection(self):
        """
        will close the connection and destroy the class
        """
        self.conn.close()
        self.destroy_instance()

    @classmethod
    def destroy_instance(cls):
        cls._instance = None



import unittest
from unittest.mock import MagicMock, patch

from tc_hivemind_backend.db.qdrant import QdrantSingleton


class TestQdrantSingleton(unittest.TestCase):
    def setUp(self):
        QdrantSingleton._QdrantSingleton__instance = None

    def test_singleton_pattern(self):
        instance1 = QdrantSingleton.get_instance()
        instance2 = QdrantSingleton.get_instance()

        self.assertIs(instance1, instance2)

        with self.assertRaises(Exception) as context:
            QdrantSingleton()
        self.assertEqual(str(context.exception), "This class is a singleton!")

    @patch("tc_hivemind_backend.db.qdrant.load_qdrant_credentials")
    @patch("tc_hivemind_backend.db.qdrant.QdrantClient")
    def test_initialization_without_api_key(
        self, mock_qdrant_client, mock_load_credentials
    ):
        mock_load_credentials.return_value = {
            "host": "test_host",
            "port": 6333,
            "api_key": "",
        }

        instance = QdrantSingleton.get_instance()

        mock_qdrant_client.assert_called_once_with(host="test_host", port=6333)

        self.assertEqual(instance.get_client(), mock_qdrant_client.return_value)

    @patch("tc_hivemind_backend.db.qdrant.load_qdrant_credentials")
    @patch("tc_hivemind_backend.db.qdrant.QdrantClient")
    def test_initialization_with_api_key(
        self, mock_qdrant_client, mock_load_credentials
    ):
        mock_load_credentials.return_value = {
            "host": "test_host",
            "port": 6333,
            "api_key": "test_api_key",
        }

        instance = QdrantSingleton.get_instance()

        mock_qdrant_client.assert_called_once_with(
            host="test_host", port=6333, api_key="test_api_key", https=False
        )

        self.assertEqual(instance.get_client(), mock_qdrant_client.return_value)

    @patch("tc_hivemind_backend.db.qdrant.load_qdrant_credentials")
    @patch("tc_hivemind_backend.db.qdrant.QdrantClient")
    @patch("logging.info")
    @patch("logging.error")
    def test_successful_connection(
        self, mock_error_log, mock_info_log, mock_qdrant_client, mock_load_credentials
    ):
        mock_load_credentials.return_value = {
            "host": "test_host",
            "port": 6333,
            "api_key": "test_api_key",
        }

        mock_client = MagicMock()
        mock_client.get_collections.return_value = []
        mock_qdrant_client.return_value = mock_client

        QdrantSingleton.get_instance()

        mock_info_log.assert_called_once_with("QDrant Connected Successfully!")
        mock_error_log.assert_not_called()

    @patch("tc_hivemind_backend.db.qdrant.load_qdrant_credentials")
    @patch("tc_hivemind_backend.db.qdrant.QdrantClient")
    @patch("logging.info")
    @patch("logging.error")
    def test_failed_connection(
        self, mock_error_log, mock_info_log, mock_qdrant_client, mock_load_credentials
    ):
        mock_load_credentials.return_value = {
            "host": "test_host",
            "port": 6333,
            "api_key": "test_api_key",
        }

        mock_client = MagicMock()
        mock_client.get_collections.side_effect = Exception("Connection failed")
        mock_qdrant_client.return_value = mock_client

        QdrantSingleton.get_instance()

        mock_error_log.assert_called_once_with(
            "QDrant not connected! exp: Connection failed"
        )
        mock_info_log.assert_not_called()

    @patch("tc_hivemind_backend.db.qdrant.load_qdrant_credentials")
    def test_credentials_loading_error(self, mock_load_credentials):
        mock_load_credentials.side_effect = Exception("Failed to load credentials")

        with self.assertRaises(Exception) as context:
            QdrantSingleton.get_instance()
        self.assertEqual(str(context.exception), "Failed to load credentials")



import logging
from datetime import datetime

from dateutil.parser import parse
from llama_index.core import Document, MockEmbedding
from llama_index.core.ingestion import (
    DocstoreStrategy,
    IngestionCache,
    IngestionPipeline,
)
from llama_index.core.node_parser import SemanticSplitterNodeParser
from llama_index.core.schema import BaseNode
from llama_index.storage.docstore.mongodb import MongoDocumentStore
from llama_index.storage.kvstore.redis import RedisKVStore as RedisCache
from qdrant_client.conversions import common_types as qdrant_types
from qdrant_client.http import models
from tc_hivemind_backend.db.credentials import Credentials
from tc_hivemind_backend.db.mongo import get_mongo_uri
from tc_hivemind_backend.db.qdrant import QdrantSingleton
from tc_hivemind_backend.db.redis import RedisSingleton
from tc_hivemind_backend.db.utils.model_hyperparams import load_model_hyperparams
from tc_hivemind_backend.embeddings.cohere import CohereEmbedding
from tc_hivemind_backend.qdrant_vector_access import QDrantVectorAccess


class CustomIngestionPipeline:
    def __init__(
        self,
        community_id: str,
        collection_name: str,
        testing: bool = False,
        use_cache: bool = True,
    ):
        self.community_id = community_id
        self.qdrant_client = QdrantSingleton.get_instance().client

        credentials = Credentials()
        _, self.embedding_dim = load_model_hyperparams()
        self.pg_creds = credentials.load_postgres()
        self.redis_cred = credentials.load_redis()
        self.collection_name = f"{community_id}_{collection_name}"
        self.platform_name = collection_name

        self.embed_model = (
            CohereEmbedding()
            if not testing
            else MockEmbedding(embed_dim=self.embedding_dim)
        )
        if use_cache:
            self.redis_client = RedisSingleton.get_instance().get_client()
        else:
            self.redis_client = None

    def run_pipeline(self, docs: list[Document]) -> list[BaseNode]:
        """
        vectorize and ingest data into a qdrant collection

        Note: This will handle duplicate documents by doing an upsert operation.

        Parameters
        ------------
        docs : list[llama_index.Document]
            list of llama-index documents

        Returns
        ---------
        nodes : list[BaseNode]
            The set of transformed and loaded Nodes/Documents
            (transformation is chunking and embedding of data)
        """
        # qdrant is just collection based and doesn't have any database
        logging.info(
            f"{len(docs)} documents were extracted and are now loading into Qdrant DB!"
        )
        vector_access = QDrantVectorAccess(collection_name=self.collection_name)
        vector_store = vector_access.setup_qdrant_vector_store()

        if self.redis_client:
            cache = IngestionCache(
                cache=RedisCache.from_redis_client(self.redis_client),
                collection=f"{self.collection_name}_ingestion_cache",
                docstore_strategy=DocstoreStrategy.UPSERTS,
            )
        else:
            cache = None

        pipeline = IngestionPipeline(
            transformations=[
                SemanticSplitterNodeParser(embed_model=self.embed_model),
                self.embed_model,
            ],
            docstore=MongoDocumentStore.from_uri(
                uri=get_mongo_uri(),
                db_name=f"docstore_{self.community_id}",
                namespace=self.platform_name,
            ),
            vector_store=vector_store,
            cache=cache,
            docstore_strategy=DocstoreStrategy.UPSERTS,
        )
        logging.info("Pipeline created, now inserting documents into pipeline!")

        nodes = pipeline.run(documents=docs, show_progress=True)
        return nodes

    def _create_payload_index(
        self,
        field_name: str,
        field_schema: qdrant_types.PayloadSchemaType,
    ) -> qdrant_types.UpdateResult:
        """
        Creates an index on a field under the payload of points in qdrant db

        Note: this could be used for payload fields that we want to scroll for after

        Parameters
        ------------
        field_name : str
            the field name under points' payload to create the index for
        field_schema : qdrant_client.conversions.common_types.PayloadSchemaType
            the schema type of the field

        Returns
        -----------
        operation_result : qdrant_client.conversions.common_types.UpdateResult
            the payload index creation type
        """
        operation_result = self.qdrant_client.create_payload_index(
            collection_name=self.collection_name,
            field_name=field_name,
            field_schema=field_schema,
        )

        return operation_result

    def get_latest_document_date(
        self,
        field_name: str,
        field_schema: qdrant_types.PayloadSchemaType = models.PayloadSchemaType.FLOAT,
    ) -> datetime | None:
        """
        get the latest date for the most recent available document

        NOTE: the given `field_name` under the points' schema MUST CONTAIN A VALUE HAVING DATE FORMAT (or a string format). If not, errors might raise in result of this function

        Parameters
        ------------
        field_name : str
            the datetime field name in qdrant points' payload
        field_schema : qdrant_client.conversions.common_types.PayloadSchemaType
            the date field schema
            for default we're assuming it is a float timestamp
            but it also could be DATETIME

        Returns
        ---------
        latest_date : datetime.datetime | None
            the datetime for the document containing the latest date
            if no document or any errors raised, we would return `None`
        """
        latest_date: datetime | None = None
        try:
            result = self._create_payload_index(
                field_name=field_name,
                field_schema=field_schema,
            )
            if result.status.name == "COMPLETED":
                latest_document = self.qdrant_client.scroll(
                    collection_name=self.collection_name,
                    limit=1,
                    with_payload=True,
                    order_by=models.OrderBy(
                        key=field_name,
                        direction=models.Direction.DESC,
                    ),
                )

                if not latest_document[0]:
                    logging.info("No documents found in the collection.")
                    latest_date = None
                else:
                    date_field = latest_document[0][0].payload[field_name]

                    # if it was float timestamp
                    if field_schema == models.PayloadSchemaType.FLOAT:
                        latest_date = datetime.fromtimestamp(date_field)

                    # it should be datetime in any other case
                    else:
                        latest_date = parse(date_field)

            else:
                raise ValueError(
                    f"Index not created successfully! index creation result: {result}"
                )
        except Exception as exp:
            logging.error(f"Error: {exp} while loading latest point!")
            latest_date = None

        return latest_date



from unittest import TestCase

from llama_index.core import MockEmbedding
from llama_index.core.indices.vector_store import VectorStoreIndex
from llama_index.vector_stores.qdrant import QdrantVectorStore
from tc_hivemind_backend.qdrant_vector_access import QDrantVectorAccess


class TestQdrantVectorAccess(TestCase):
    def test_init(self):
        expected_collection_name = "sample_collection"
        qdrant_vector_access = QDrantVectorAccess(
            collection_name=expected_collection_name,
            testing=True,
        )

        self.assertEqual(qdrant_vector_access.collection_name, expected_collection_name)
        self.assertIsInstance(qdrant_vector_access.embed_model, MockEmbedding)

    def test_setup_index(self):
        qdrant_vector_access = QDrantVectorAccess(
            collection_name="sample_collection",
            testing=True,
        )
        vector_store = qdrant_vector_access.setup_qdrant_vector_store()
        self.assertIsInstance(vector_store, QdrantVectorStore)

    def test_load_index(self):
        qdrant_vector_access = QDrantVectorAccess(
            collection_name="sample_collection",
            testing=True,
        )
        index = qdrant_vector_access.load_index()
        self.assertIsInstance(index, VectorStoreIndex)



from datetime import datetime, timedelta
from unittest import TestCase

from bson import ObjectId
from tc_hivemind_backend.db.modules_base import ModulesBase
from tc_hivemind_backend.db.mongo import MongoSingleton


class TestModulesBaseQueryToken(TestCase):
    def setUp(self) -> None:
        self.client = MongoSingleton.get_instance().get_client()
        self.client["Core"].drop_collection("tokens")
        self.client["Core"].drop_collection("platforms")

    def test_one_token(self):
        sample_user = ObjectId("5d7baf326c8a2e2400000000")
        community_id = ObjectId("6579c364f1120850414e0dc5")
        sample_token_type = "type1"
        sample_token_value = "tokenid12345"
        platform_id = ObjectId("6579c364f1120850414e0dc6")

        self.client["Core"]["platforms"].insert_one(
            {
                "_id": platform_id,
                "name": "platform_name",
                "metadata": {
                    "id": "113445975232201081511",
                    "userId": str(sample_user),
                },
                "community": community_id,
                "disconnectedAt": None,
                "connectedAt": datetime.now(),
                "createdAt": datetime.now(),
                "updatedAt": datetime.now(),
            }
        )

        sample_token_doc = {
            "token": sample_token_value,
            "user": sample_user,
            "type": sample_token_type,
            "expires": datetime.now() + timedelta(days=1),
            "blacklisted": False,
            "createdAt": datetime.now() - timedelta(days=1),
            "updatedAt": datetime.now() - timedelta(days=1),
        }
        self.client["Core"]["tokens"].insert_one(sample_token_doc)
        token = ModulesBase().get_token(
            platform_id=platform_id, token_type=sample_token_type
        )

        self.assertEqual(token, sample_token_value)

    def test_empty_tokens_collection(self):
        platform_id = ObjectId("6579c364f1120850414e0dc6")
        sample_token_type = "type1"
        with self.assertRaises(ValueError):
            _ = ModulesBase().get_token(
                platform_id=platform_id, token_type=sample_token_type
            )

    def test_no_platform(self):
        sample_user = ObjectId("5d7baf326c8a2e2400000000")
        platform_id = ObjectId("6579c364f1120850414e0dc6")
        sample_token_type = "type1"
        sample_token_value = "tokenid12345"

        sample_token_doc = {
            "token": sample_token_value,
            "user": sample_user,
            "type": sample_token_type,
            "expires": datetime.now() + timedelta(days=1),
            "blacklisted": False,
            "createdAt": datetime.now() - timedelta(days=1),
            "updatedAt": datetime.now() - timedelta(days=1),
        }
        self.client["Core"]["tokens"].insert_one(sample_token_doc)
        with self.assertRaises(ValueError):
            _ = ModulesBase().get_token(
                platform_id=platform_id, token_type=sample_token_type
            )

    def test_no_token(self):
        sample_user = ObjectId("5d7baf326c8a2e2400000000")
        sample_user_with_no_token = ObjectId("5d7baf326c8a2e2400000001")

        platform_id = ObjectId("6579c364f1120850414e0dc6")
        sample_token_type = "type1"
        sample_token_value = "tokenid12345"
        community_id = ObjectId("6579c364f1120850414e0dc5")

        self.client["Core"]["platforms"].insert_one(
            {
                "_id": platform_id,
                "name": "platform_name",
                "metadata": {
                    "id": "113445975232201081511",
                    "userId": str(sample_user_with_no_token),
                },
                "community": community_id,
                "disconnectedAt": None,
                "connectedAt": datetime.now(),
                "createdAt": datetime.now(),
                "updatedAt": datetime.now(),
            }
        )

        sample_token_doc = {
            "token": sample_token_value,
            "user": sample_user,
            "type": sample_token_type,
            "expires": datetime.now() + timedelta(days=1),
            "blacklisted": False,
            "createdAt": datetime.now() - timedelta(days=1),
            "updatedAt": datetime.now() - timedelta(days=1),
        }
        self.client["Core"]["tokens"].insert_one(sample_token_doc)
        with self.assertRaises(ValueError):
            _ = ModulesBase().get_token(
                platform_id=platform_id, token_type=sample_token_type
            )



import os

from dotenv import load_dotenv


def load_model_hyperparams() -> tuple[int, int]:
    """
    load the llm and embedding model hyperparameters (the input parameters)

    Returns
    ---------
    chunk_size : int
        the chunk size to chunk the data
    embedding_dim : int
        the embedding dimension
    """
    load_dotenv()

    chunk_size_str = os.getenv("CHUNK_SIZE")
    chunk_size: int
    if chunk_size_str is None:
        raise ValueError("Chunk size is not given in env")
    else:
        chunk_size = int(chunk_size_str)

    embedding_dim_str = os.getenv("EMBEDDING_DIM")
    embedding_dim: int
    if embedding_dim_str is None:
        raise ValueError("Embedding dimension size is not given in env")
    else:
        embedding_dim = int(embedding_dim_str)

    return chunk_size, embedding_dim



import unittest
from unittest.mock import Mock

from llama_index.core.ingestion import IngestionPipeline
from llama_index.core.schema import Document
from tc_hivemind_backend.ingest_qdrant import CustomIngestionPipeline


class TestIngestionPipeline(unittest.TestCase):
    def test_run_pipeline(self):
        ingest_pipeline = Mock(IngestionPipeline)
        community = "1234"
        collection_name = "google"
        ingestion_pipeline = CustomIngestionPipeline(
            community_id=community, collection_name=collection_name, testing=True
        )
        docs = [
            Document(
                id_="b049e7cf-3279-404b-b324-9776fe1cf60b",
                text="""A banana is an elongated""",
            ),
            Document(
                id_="3b3033c0-7e37-493c-8b4c-fd51f754a59a",
                text="""Musa species are native to tropical Indomalaya and Australia""",
            ),
        ]

        processed_result = ingestion_pipeline.run_pipeline(docs)
        ingest_pipeline.run.return_value = processed_result
        self.assertEqual(len(processed_result), 2)

    def test_load_pipeline_run_exception(self):
        ingestion_pipeline = CustomIngestionPipeline(
            "1234", collection_name="google", testing=True
        )
        ingestion_pipeline.run_pipeline = Mock()
        ingestion_pipeline.run_pipeline.side_effect = Exception("Test Exception")
        docs = ["ww"]
        with self.assertRaises(Exception) as context:
            ingestion_pipeline.run_pipeline(docs)
        self.assertEqual(str(context.exception), "Test Exception")
        ingestion_pipeline.run_pipeline.assert_called_with(docs)



import unittest
from unittest.mock import MagicMock, patch

from tc_hivemind_backend.db.redis import RedisSingleton


class TestRedisSingleton(unittest.TestCase):
    def setUp(self):
        RedisSingleton._RedisSingleton__instance = None

    def test_singleton_pattern(self):
        instance1 = RedisSingleton.get_instance()
        instance2 = RedisSingleton.get_instance()

        self.assertIs(instance1, instance2)

        with self.assertRaises(Exception) as context:
            RedisSingleton()
        self.assertEqual(str(context.exception), "This class is a singleton!")

    @patch("tc_hivemind_backend.db.redis.Credentials")
    @patch("redis.Redis")
    def test_initialization(self, mock_redis, mock_credentials):
        mock_credentials_instance = MagicMock()
        mock_credentials_instance.load_redis.return_value = {
            "host": "test_host",
            "port": "6379",
            "password": "test_password",
        }
        mock_credentials.return_value = mock_credentials_instance

        instance = RedisSingleton.get_instance()

        mock_redis.assert_called_once_with(
            host="test_host", port=6379, password="test_password", decode_responses=True
        )

        self.assertEqual(instance.get_client(), mock_redis.return_value)

    @patch("tc_hivemind_backend.db.redis.Credentials")
    @patch("redis.Redis")
    @patch("logging.info")
    @patch("logging.error")
    def test_successful_connection(
        self, mock_error_log, mock_info_log, mock_redis, mock_credentials
    ):
        mock_credentials_instance = MagicMock()
        mock_credentials_instance.load_redis.return_value = {
            "host": "test_host",
            "port": "6379",
            "password": "test_password",
        }
        mock_credentials.return_value = mock_credentials_instance

        mock_client = MagicMock()
        mock_client.ping.return_value = True
        mock_redis.return_value = mock_client

        RedisSingleton.get_instance()

        mock_info_log.assert_called_once_with(
            "Redis Connected Successfully! Ping returned: True"
        )
        mock_error_log.assert_not_called()

    @patch("tc_hivemind_backend.db.redis.Credentials")
    @patch("redis.Redis")
    @patch("logging.info")
    @patch("logging.error")
    def test_failed_connection(
        self, mock_error_log, mock_info_log, mock_redis, mock_credentials
    ):
        mock_credentials_instance = MagicMock()
        mock_credentials_instance.load_redis.return_value = {
            "host": "test_host",
            "port": "6379",
            "password": "test_password",
        }
        mock_credentials.return_value = mock_credentials_instance

        mock_client = MagicMock()
        mock_client.ping.side_effect = Exception("Connection failed")
        mock_redis.return_value = mock_client

        RedisSingleton.get_instance()

        mock_error_log.assert_called_once_with(
            "Redis not connected! exp: Connection failed"
        )
        mock_info_log.assert_not_called()

    @patch("tc_hivemind_backend.db.redis.Credentials")
    def test_credentials_loading_error(self, mock_credentials):
        mock_credentials_instance = MagicMock()
        mock_credentials_instance.load_redis.side_effect = Exception(
            "Failed to load credentials"
        )
        mock_credentials.return_value = mock_credentials_instance

        with self.assertRaises(Exception) as context:
            RedisSingleton.get_instance()
        self.assertEqual(str(context.exception), "Failed to load credentials")


if __name__ == "__main__":
    unittest.main()
